/* SampEn_keepNA.c:
* The codes are built on mse.c from https://urldefense.proofpoint.com/v2/url?u=https-3A__www.physionet.org_physiotools_mse_mse.c&d=BQIGAg&c=KXXihdR8fRNGFkKiMQzstu-8MbOxd1NuZkcSBymGmgo&r=TcM8b82z3tegOJXIRc2Vz7clBmJyyKx9Zj12bbp3ukQ&m=4nLbigf5_1qJiGHdFI8ZiGDhvou2n5icBydzoa9wGqA&s=blB40K2y5Alb2amu_G4HOWJ1LecEzeB5MZJ2Xn_YsTQ&e= .
* The improvement is in handling with missing value. mse.c does not allow the input data
*  to have missing values whereas SampEn_keepNA.c does not only allow the input data to
*  missing value but also contains codes to handle with missing values in the calculation
*  process.
* This code only contains the part of sample entropy calculation.
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <float.h>


#define MAXSTR 1250     /* maximum string size */
#define DATA_MAX 200000 /* maximum number of data points */
#define M_MAX 10        /* maximum value of the parameter m */
#define FILEMAX 100     /* maximum number of data files in the list file*/

/* Global variables */
char *prog, line[MAXSTR];
double SE[FILEMAX][M_MAX];
double *u;
int nlin,  m_max, i_min, i_max;

static char file[500];

FILE *fl;
FILE *pin;

main(argc, argv)
int argc;
char *argv[];
{

    int i, j, k, l, m, nfile, flag;
    double r, sd, tmp;

    /* Function prototypes */
    void ReadData(void);
    double *array(int n);
    double StandardDeviation(void);
    void SampleEntropy (int ll, double r, double sd);
    void PrintResults (int nfile);

    /* Initialize variables. */
    prog = argv[0];
    m_max = 2;
    i_min = 0;
    i_max = 39999;
    nfile = 0;
    flag = 0;


    /* Read and interpret the command line. */
    i = 0;
    while (++i < argc && *argv[i] == '-') {
	switch(argv[i][1]) {

	  case 'F':
	    if ((fl = fopen(argv[++i], "r")) == NULL) {
		fprintf(stderr, "%s [-F]: can't open input file %s\n",
			prog, argv[i]);
		exit(1);
	    }
	    flag=1;
	    break;

	  case 'i':
	    if ((i_min = atoi(argv[++i])) < 0) {
		fprintf(stderr,
	           "%s [-i]: minimum i must not be less than 0 (default: 0)\n",
			prog);
		exit(1);
	    }
	    break;

	  case 'I':
	    if ((i_max = atoi(argv[++i])) <= 0 || i_max <= i_min) {
		fprintf(stderr,
			"%s [-I]: maximum i must be greater than %d "
			"(default: number of data points)\n",
			prog, i_min);
		exit(1);
	    }
	    break;


	  default:
	    usage();
	    exit(1);
	}
    }

    /* Memory allocation. */
    u = array(DATA_MAX);


    /* Process a single data file. */
    if (flag == 0) {
	l = 0;
	nfile = 1;

	/* Read data from stdin. */
	pin = stdin;
	ReadData();

	/* Calculate standard deviation. */
	sd = StandardDeviation();

    SampleEntropy(l, 0.15, sd);

	/* Print results. */
	PrintResults(nfile);
    }


    /* Process multiple data files. */
    if (flag == 1) {

	/* Read the list of data files. */
	for (l = 0; fscanf(fl, "%s", file) == 1; l++) {
	    nfile++;   /*count the number of data files*/

	    if ((pin = fopen(file, "r")) == NULL) {   /* open each data file */
		fprintf(stderr, "%s : Cannot open file %s\n", prog, file);
		exit(1);
	    }


	    /* Read the data. */
	    ReadData();

	    /* Calculate the standard deviation. */
	    sd = StandardDeviation ();

        SampleEntropy (l, 0.15, sd);
	}

	/* Print results. */
	PrintResults(nfile);

	fclose(pin);
	fclose(fl);
    }

    free(u);
    exit(0);
}

double *array(int n)
{
    int i;
    double *a;

    if ((a = calloc (n, sizeof (double))) == NULL) {
	fprintf(stderr, "%s : insufficient memory\n", prog);
	exit(2);
    }
    return (a);
}

void ReadData(void)
{

    nlin=0;
    int j = -1;

    int nmissing=0;


    while (fgets(line, MAXSTR, pin) != NULL) {
	j++;
	if (j >= i_min && j <= i_max) {
	    if(sscanf(line, "%lf", &u[nlin])==0) //sscanf return 0 if no valid number found in line
        {
            u[nlin] = DBL_MAX;
            nmissing++;
        }
	    nlin++;
	}
    }

}

double StandardDeviation(void)
{
    double sum=0.0, sum2=0.0, sd;
    int j;

    int count=0;

    for (j = 0; j < nlin; j++) {
        if(u[j]==DBL_MAX)
        {
            ;
        }
        else
        {
            sum += u[j];
            sum2 += u[j] * u[j];
            count++;
        }
    }
    sd = sqrt((sum2 - sum*sum/count)/(count - 1));
    return (sd);
}


void SampleEntropy(int ll, double r, double sd)
{
    int i, k, l, nlin_j, kk;
    int cont[M_MAX+1];
    double r_new;

    nlin_j = nlin - m_max;

    // It's meaningless to compare when you have only one point
    if(nlin_j<2)
        return;
    r_new = r*sd;

    for (i = 0; i < M_MAX; i++)
	cont[i]=0;

    for (i = 0; i < nlin_j; ++i) {
	for (l = i+1; l < nlin_j; ++l) { /*self-matches are not counted*/
	    k = 0;
	    while (k <= m_max && u[i+k]!=DBL_MAX && u[l+k]!=DBL_MAX)
            k++;

        if(k==m_max+1)
        {
            kk = 0;
            while(kk <= m_max && fabs(u[i+kk] - u[l+kk]) <= r_new)
                cont[++kk]++;
        }
	}
    }


    for (i = 1; i <= m_max; i++)

        if (cont[i+1] == 0 || cont[i] == 0)
            SE[ll][i] = -log((double)1/((nlin_j)*(nlin_j-1)));
        else
            SE[ll][i] = -log((double)cont[i+1]/cont[i]);

}



void PrintResults(int nfile)
{
    int l;

    if (nfile > 1) {
    fseek(fl, 0, SEEK_SET);
    for(l = 0; fscanf(fl, "%s", file) == 1; l++)
        printf("%.6s\t", file);
    printf("\n");
    }
    for (l=0; l<nfile; l++)
        printf("%.3lf\t", SE[l][2]);
    printf("\n");
}


usage()
{
    fprintf(stderr, "usage: %s [options]\n", prog);
    fprintf(stderr, "\nTo calculate SE for a single data file:\n"
	    "    %s <datafile >outputfile\n"
	    "To calculate SE for multiple data files:\n"
	    "    %s -F listfile >outputfile\n"
	    "(where listfile contains a list of data files).\n\n", prog, prog);
    fprintf(stderr, "Data files should contain a single column of numbers\n");
    fprintf(stderr, "Options may include:\n");

    fprintf(stderr, "  -i N  set minimum i to N [0-39998; default: 0]\n");
    fprintf(stderr, "  -I N   set maximum i to N [1-39999: default: 39999]\n");

	fprintf(stderr,
 "Option arguments indicated as N are integers. \n");
}
